from dataclasses import dataclass

from tdm.abstract.datamodel import AbstractLinkDomainType, Identifiable
from ._composite import CompositeValueType
from ._property import ValueType


@dataclass(frozen=True)
class ComponentValueType(Identifiable, AbstractLinkDomainType[CompositeValueType, ValueType]):
    """
    Domain type for component facts
    """
    pass



from dataclasses import dataclass, field, replace
from typing import Dict, Iterable, List, Optional, Set


@dataclass(frozen=True)
class NodesStructure(object):
    nodes: Set[str] = field(default_factory=set)
    roots: Set[str] = field(default_factory=set)
    children: Dict[str, List[str]] = field(default_factory=dict)
    parent: Dict[str, str] = field(default_factory=dict)
    main_root: Optional[str] = None

    def with_nodes(self, nodes: Set[str]) -> 'NodesStructure':
        if not nodes:
            return self
        new_nodes = nodes.difference(self.nodes)
        roots = self.roots.union(new_nodes)
        return replace(self, nodes=self.nodes.union(new_nodes), roots=roots)

    def remove_ids(self, ids: Set[str]) -> 'NodesStructure':
        if not ids:
            return self
        ids = ids.intersection(self.nodes)  # ignore excess ids
        nodes = self.nodes.difference(ids)
        roots = self.roots.difference(ids)
        parent = dict(self.parent)
        children = dict(self.children)
        main_root = None if self.main_root in ids else self.main_root
        updated = set()
        for id_ in ids:
            parent_id = parent.pop(id_, None)
            if parent_id is not None:
                _copy_if_not_updated(children, parent_id, updated).remove(id_)
            for child_id in children.pop(id_, ()):
                parent.pop(child_id)
                roots.add(child_id)
        return NodesStructure(nodes, roots, _prune(children, updated), parent, main_root)

    def update_nodes(self, nodes: Set[str]) -> 'NodesStructure':
        removed = self.nodes.difference(nodes)
        added = nodes.difference(self.nodes)
        return self.remove_ids(removed).with_nodes(added)

    def with_main_root(self, main_root: Optional[str]) -> 'NodesStructure':
        if main_root is None or main_root in self.roots:
            return replace(self, main_root=main_root)
        raise ValueError(f"node {main_root} is not one of the roots")

    def with_children(self, links: Dict[str, Iterable[str]], *, force: bool = False) -> 'NodesStructure':
        roots: Set[str] = set(self.roots)
        parents: Dict[str, str] = dict(self.parent)  # child -> parent
        children: Dict[str, List[str]] = dict(self.children)  # parent -> [child]
        updated: Set[str] = set()
        seed: Set[str] = set()  # added parents to control loops in graph

        for parent_id, children_ids in links.items():
            if parent_id not in self.nodes:
                raise ValueError(f"node {parent_id} not contained in the document")
            children_ids = list(children_ids)
            for child_id in children_ids:
                if child_id not in self.nodes:
                    raise ValueError(f"node {child_id} not contained in the document")
            roots.difference_update(children_ids)

            if not force:
                # check if some node has another parent
                for child_id in set(children_ids).intersection(parents):
                    if parents[child_id] != parent_id:
                        raise ValueError(f"couldn't link {child_id} with {parent_id} as it already linked with {parents[child_id]}")

            # update children collection
            old_children = _copy_if_not_updated(children, parent_id, updated)
            intersection = set(old_children).intersection(children_ids)
            if force:
                # remove intersecting ids from old collection
                for child_id in intersection:
                    old_children.remove(child_id)
            else:
                # remove intersecting ids from new collection
                children_ids = [child_id for child_id in children_ids if child_id not in intersection]
            old_children.extend(children_ids)

            # update parents collection
            for child_id in children_ids:
                old_parent = parents.get(child_id)
                if old_parent is not None:  # child had parent, so we should remove link from parent collection
                    if old_parent == parent_id:
                        continue
                    _copy_if_not_updated(children, old_parent, updated).remove(child_id)
                parents[child_id] = parent_id
                seed.add(child_id)

        _check_circle(parents, seed)

        if not force and self.main_root is not None and self.main_root not in roots:
            raise ValueError(f"main root {self.main_root} couldn't be child node")

        return NodesStructure(self.nodes, roots, _prune(children, updated), parents, self.main_root if self.main_root in roots else None)

    def with_parent(self, parent: str, child: str, *, force: bool = False) -> 'NodesStructure':
        return self.with_children({parent: (child,)}, force=force)

    def without_children(self, links: Dict[str, Set[str]]) -> 'NodesStructure':
        roots = set(self.roots)
        parents: Dict[str, str] = dict(self.parent)  # child -> parent
        children: Dict[str, List[str]] = dict(self.children)  # parent -> [child]
        updated: Set[str] = set()

        for parent_id, children_ids in links.items():
            if parent_id not in children:  # nothing to remove
                continue
            children_ids = children_ids.intersection(children[parent_id])
            children[parent_id] = [child_id for child_id in children[parent_id] if child_id not in children_ids]
            updated.add(parent_id)
            for child_id in self.nodes.intersection(children_ids):
                parents.pop(child_id, None)
                roots.add(child_id)

        return replace(self, roots=roots, parent=parents, children=_prune(children, updated))

    def as_roots(self, nodes: Iterable[str]) -> 'NodesStructure':
        nodes = self.nodes.intersection(nodes)
        roots = self.roots.union(nodes)
        parents: Dict[str, str] = dict(self.parent)  # child -> parent
        children: Dict[str, List[str]] = dict(self.children)  # parent -> [child]
        updated: Set[str] = set()

        for child_id in nodes:
            parent_id = parents.pop(child_id, None)
            if parent_id is not None:
                _copy_if_not_updated(children, parent_id, updated).remove(child_id)

        return replace(self, roots=roots, parent=parents, children=_prune(children, updated))


def _copy_if_not_updated(children: Dict[str, List[str]], key: str, updated: Set[str]) -> List[str]:
    if key not in updated:
        children[key] = list(children.get(key, ()))
        updated.add(key)
    return children[key]


def _prune(children: Dict[str, List[str]], updated: Set[str]) -> Dict[str, List[str]]:
    for key in updated:
        if not children.get(key):
            children.pop(key, None)
    return children


def _check_circle(graph: Dict[str, str], seed: Set[str]) -> None:
    state = {key: graph.get(key) for key in seed}
    while state:
        s = {}
        for key, obj in state.items():
            if obj is None:
                continue
            if key == obj:
                raise ValueError(f"circle dependency with element {key}")
            s[key] = graph.get(obj)
        state = s



from enum import Enum
from typing import Union

from pydantic import BaseModel
from typing_extensions import Literal


class DirectiveType(str, Enum):
    CREATE_ACCOUNT = 'create_account'
    CREATE_PLATFORM = 'create_platform'


class AbstractDirectiveModel(BaseModel):
    directive_type: DirectiveType


class CreateAccountDirectiveModel(AbstractDirectiveModel):
    key: str
    platform_key: str
    name: str
    url: str

    directive_type: Literal[DirectiveType.CREATE_ACCOUNT] = DirectiveType.CREATE_ACCOUNT

    def __hash__(self):
        return hash((self.directive_type, self.key))


class CreatePlatformDirectiveModel(AbstractDirectiveModel):
    key: str
    platform_type: str
    name: str
    url: str

    directive_type: Literal[DirectiveType.CREATE_PLATFORM] = DirectiveType.CREATE_PLATFORM

    def __hash__(self):
        return hash((self.directive_type, self.key))


DirectiveModel = Union[CreateAccountDirectiveModel, CreatePlatformDirectiveModel]



from abc import ABCMeta, abstractmethod
from typing import Callable, Hashable, Iterable, Optional, Tuple

from .identifiable import EnsureIdentifiable


class Duplicatable(metaclass=ABCMeta):
    @abstractmethod
    def duplicate_hash(self) -> Hashable:
        pass

    @abstractmethod
    def choose_one(
            self, other, related: Callable[[str], Iterable[str]], restore_element: Callable[[str], EnsureIdentifiable]
    ) -> Optional[Tuple[str, str]]:
        pass



__all__ = [
    'AbstractFact', 'FactStatus',
    'AbstractLinkFact',
    'AbstractValueFact'
]

from ._fact import AbstractFact, FactStatus
from ._link import AbstractLinkFact
from ._value import AbstractValueFact



from abc import abstractmethod
from dataclasses import dataclass, field
from typing import Generic, Type, TypeVar

from typing_extensions import Self

from tdm.abstract.datamodel import AbstractMarkup, AbstractNode
from ._interface import AbstractNodeWrapper

_Node = TypeVar('_Node', bound=AbstractNode)
_Markup = TypeVar('_Markup', bound=AbstractMarkup)


@dataclass(frozen=True)
class AbstractNodeWrapperImpl(AbstractNodeWrapper[_Node], Generic[_Node, _Markup]):
    markup: _Markup = field(default_factory=lambda: None)

    def _convert_markup(self, markup: AbstractMarkup) -> _Markup:
        markup_type: Type[_Markup] = self._markup_type()
        if isinstance(markup, markup_type):
            return markup
        return markup_type.from_markup(markup)

    @classmethod
    def wrap(cls, node: _Node) -> Self:
        if isinstance(node, cls):
            return node  # do nothing
        if not isinstance(node, cls._node_type()):
            raise ValueError  # wrap only appropriate nodes
        args = dict(node.__dict__)
        args['markup'] = cls._markup_type().from_markup(node.markup)
        return cls(**args)  # replace markup

    @classmethod
    @abstractmethod
    def _node_type(cls) -> Type[_Node]:
        pass

    @classmethod
    @abstractmethod
    def _markup_type(cls) -> Type[_Markup]:
        pass

    def __eq__(self, other):
        node_type = self._node_type()
        if not isinstance(other, node_type):
            return NotImplemented
        return self.__dict__ == other.__dict__



__all__ = [
    'ReferenceNodeLink', 'TranslationNodeLink', 'EquivalenceNodeLink'
]

from .eqivalence import EquivalenceNodeLink
from .reference import ReferenceNodeLink
from .translation import TranslationNodeLink



from dataclasses import dataclass
from typing import Callable, Optional

from tdm.abstract.datamodel import AbstractNodeLink, AbstractNodeMention, Identifiable
from tdm.abstract.json_schema import generate_model


@generate_model(label='translation')
@dataclass(frozen=True)
class TranslationNodeLink(Identifiable, AbstractNodeLink[AbstractNodeMention, AbstractNodeMention]):
    """
    Represents a link between two node mentions where one is a translation of the other.

    Attributes
    --------
    language:
        The language of the translation. Default is None.
    """

    language: Optional[str] = None

    @staticmethod
    def lang_filter(language: str) -> Callable[['TranslationNodeLink'], bool]:
        """
        Returns a filter function based on the translation language
        :param language: The language to filter by
        :return: The filter function for TranslationNodeLink objects
        """

        def _filter(link: TranslationNodeLink) -> bool:
            return link.language == language

        return _filter



from dataclasses import dataclass
from typing import Optional

from tdm.abstract.datamodel import AbstractContentNode, BaseNodeMetadata
from tdm.abstract.json_schema import generate_model


@dataclass(frozen=True)
class FileNodeMetadata(BaseNodeMetadata):
    """
    File node metadata

    Attributes
    --------
    name:
        file name
    size:
        file size (bytes)
    """
    name: Optional[str] = None
    size: Optional[int] = None


@generate_model(label='file')
@dataclass(frozen=True)
class FileNode(AbstractContentNode[FileNodeMetadata, str]):
    """
    Node for file representation.
    ``content`` contains file URI.
    """
    pass


@dataclass(frozen=True)
class ImageNodeMetadata(FileNodeMetadata):
    """
    Image node metadata

    Attributes
    --------
    width:
        image width
    height:
        image height
    """
    width: Optional[int] = None
    height: Optional[int] = None


@generate_model(label='image')
@dataclass(frozen=True)
class ImageNode(AbstractContentNode[ImageNodeMetadata, str]):
    """
    Node for image representation.
    ``content`` contains image file URI.
    """
    pass



from abc import ABCMeta, abstractmethod
from typing import Callable, Dict, Iterable, Iterator, Type, TypeVar, Union

from tdm.abstract.datamodel.identifiable import Identifiable
from ._type import AbstractDomainType

_DomainType = TypeVar('_DomainType', bound=AbstractDomainType)


class AbstractDomain(metaclass=ABCMeta):
    """
    Represents an abstract domain in the knowledge base.

    This is an abstract base class that provides the basic structure and methods for a domain.

    Attributes:
        id2type (Dict[str, AbstractDomainType]): A dictionary mapping IDs to domain types.
        types (Dict[Type[AbstractDomainType], Iterable[AbstractDomainType]]): A dictionary mapping domain types to their instances.

    """
    __slots__ = ()

    @property
    @abstractmethod
    def id2type(self) -> Dict[str, AbstractDomainType]:
        """
        Provides a dictionary mapping identifiers to domain types.

        :return: A dictionary mapping identifiers to domain types.
        """
        pass

    @property
    @abstractmethod
    def types(self) -> Dict[Type[AbstractDomainType], Iterable[AbstractDomainType]]:
        """
        Provides a dictionary mapping domain types to their instances.

        :return: A dictionary mapping domain types to their instances.
        """
        pass

    @abstractmethod
    def get_type(self, id_: str) -> AbstractDomainType:
        """
        Retrieves a domain type by its identifier.

        :param id_: The identifier of the domain type.
        :return: The domain type associated with the provided identifier.

        """
        pass

    @abstractmethod
    def get_types(
            self, type_: Type[_DomainType] = AbstractDomainType, *,
            filter_: Union[Callable[[_DomainType], bool], Iterable[Callable[[_DomainType], bool]]] = tuple()
    ) -> Iterator[_DomainType]:
        """
        Retrieves instances of domain types based on the provided filters.

        :param type_: The type of domain types to retrieve. Defaults to AbstractDomainType (all domain types).
        :param filter_: A filter or a list of filters to apply on each individual retrieved domain types.
        :return: An iterator yielding instances of domain types.
        """
        pass

    @abstractmethod
    def related_types(
            self, obj: Union[Identifiable, str], type_: Type[_DomainType] = AbstractDomainType, *,
            filter_: Union[Callable[[_DomainType], bool], Iterable[Callable[[_DomainType], bool]]] = tuple()
    ) -> Iterator[_DomainType]:
        """
        Retrieves related domain types based on the provided object and filters.

        :param obj: An object or its identifier to retrieve related domain types for.
        :param type_: The type of domain types to retrieve. Defaults to AbstractDomainType (all domain types).
        :param filter_: A filter or a list of filters to apply on each individual retrieved domain types.

        :return: An iterator yielding instances of related domain types.
        """
        pass



from dataclasses import dataclass, field
from typing import ClassVar, Optional

from tdm.abstract.datamodel import AbstractContentNode, BaseNodeMetadata
from tdm.abstract.json_schema import generate_model


@dataclass(frozen=True)
class TextNodeMetadata(BaseNodeMetadata):
    """
    Text node metadata

    Attributes
    --------
    language:
        node content language
    header:
        header indicator
    header_level:
        level of the header (could be not None if ``header`` is True)
    """

    language: Optional[str] = None
    header: bool = False
    header_level: Optional[int] = None

    UNKNOWN_LANG: ClassVar[str] = field(default='unknown', init=False, hash=False, repr=False, compare=False)

    def __post_init__(self):
        if self.header_level is not None:
            if self.header_level < 0:
                raise ValueError(f"header_level could not be negative (actual: {self.header_level})")
            if not self.header:
                raise ValueError("header_level could be specified only with header flag")


@generate_model(label='text')
@dataclass(frozen=True)
class TextNode(AbstractContentNode[TextNodeMetadata, str]):
    """
    Text node implementation.
    ``content`` contains node text.
    """
    pass


@dataclass(frozen=True)
class KeyNodeMetadata(BaseNodeMetadata):
    """
    Key node metadata

    Attributes
    --------
    language:
        node content language
    """
    language: Optional[str] = None


@generate_model(label='key')
@dataclass(frozen=True)
class KeyNode(AbstractContentNode[TextNodeMetadata, str]):
    """
    Node for key representation in key-value pair.
    ``content`` contains key text.

    KeyNode could be direct child only for a ``JSONNode``.
    """
    pass



from abc import ABCMeta
from dataclasses import dataclass, replace
from typing import Callable, Hashable, Iterable, Iterator, Optional, Set, Tuple, Union

from tdm.abstract.datamodel import AbstractDomain, AbstractFact, AbstractLinkFact, EnsureIdentifiable, FactStatus, Identifiable
from tdm.abstract.datamodel.duplicatable import Duplicatable
from tdm.abstract.datamodel.fact import AbstractValueFact
from tdm.abstract.datamodel.value import AbstractValue
from tdm.abstract.json_schema import generate_model
from tdm.datamodel.common import PrunableMixin, ViewContainer
from tdm.datamodel.domain import AtomValueType, ComponentValueType, CompositeValueType


@generate_model(label='atom')
@dataclass(frozen=True)
class AtomValueFact(Identifiable, AbstractValueFact[AtomValueType, AbstractValue], PrunableMixin):
    """
    Represents a fact that holds an atomic value.

    If the fact status is FactStatus.NEW `value` should be a tuple.
    If the fact status is FactStatus.APPROVED `value` should be single-valued.
    """

    def replace_with_domain(self, domain: AbstractDomain) -> 'AtomValueFact':
        if isinstance(self.type_id, str):
            domain_type = domain.get_type(self.type_id)
            if not isinstance(domain_type, AtomValueType):
                raise ValueError
            return replace(self, type_id=domain_type)
        return self

    def _as_tuple(self) -> tuple:
        return self.id, self.str_type_id, self.value

    def __eq__(self, other):
        if not isinstance(other, AtomValueFact):
            return NotImplemented
        return self._as_tuple() == other._as_tuple()

    def __hash__(self):
        return hash(self._as_tuple())

    def _collect_links(self, links: Iterable) -> Iterator[Hashable]:
        from tdm.datamodel.facts import MentionFact

        for m in links:
            if isinstance(m, MentionFact):
                yield m.mention

    @staticmethod
    def empty_value_filter() -> Callable[['AtomValueFact'], bool]:
        """
        Create a filter function to filter `AtomValueFact` instances with empty values.

        :return: A filter function for `AtomValueFact` instances with empty values.
        """
        return lambda f: isinstance(f.value, tuple) and not f.value

    @staticmethod
    def tuple_value_filter() -> Callable[['AtomValueFact'], bool]:
        """
        Create a filter function to filter `AtomValueFact` instances with tuple values.

        :return: A filter function for `AtomValueFact` instances with tuple values.
        """
        return lambda f: isinstance(f.value, tuple)

    @staticmethod
    def single_value_filter() -> Callable[['AtomValueFact'], bool]:
        """
        Create a filter function to filter `AtomValueFact` instances with a single value.

        :return: A filter function for `AtomValueFact` instances with a single value.
        """
        return lambda f: isinstance(f.value, AbstractValue)

    def is_hanging(self, doc: ViewContainer) -> bool:
        from .mention import MentionFact
        return self.status is not FactStatus.APPROVED \
            and not self.value \
            and not tuple(doc.related_elements(AbstractFact, self, MentionFact, tuple()))


@dataclass(frozen=True)
class _CompositeValueFact(AbstractFact, metaclass=ABCMeta):
    """
    Auxiliary class for `CompositeValueFact` to fix dataclass fields order.

    Attributes
    ----------
    type_id:
        The type identifier or domain composite value type of the composite value fact.
    """
    type_id: Union[str, CompositeValueType]

    @classmethod
    def constant_fields(cls) -> Set[str]:
        return {'type_id'}

    @property
    def str_type_id(self) -> str:
        return self.type_id if isinstance(self.type_id, str) else self.type_id.id


@generate_model(label='composite')
@dataclass(frozen=True)
class CompositeValueFact(Identifiable, _CompositeValueFact, PrunableMixin, Duplicatable):
    """
    Represents a fact that holds a composite value.

    The whole value is represented as `CompositeValueFact` with related `ComponentFact`s.
    """

    def replace_with_domain(self, domain: AbstractDomain) -> 'CompositeValueFact':
        if isinstance(self.type_id, str):
            domain_type = domain.get_type(self.type_id)
            if not isinstance(domain_type, CompositeValueType):
                raise ValueError
            return replace(self, type_id=domain_type)
        return self

    def _as_tuple(self) -> tuple:
        return self.id, self.str_type_id

    def __eq__(self, other):
        if not isinstance(other, CompositeValueFact):
            return NotImplemented
        return self._as_tuple() == other._as_tuple()

    def __hash__(self):
        return hash(self._as_tuple())

    def is_hanging(self, doc: ViewContainer) -> bool:
        return self.status is not FactStatus.APPROVED and \
            not [x.target for x in doc.related_elements(AbstractFact, self, ComponentFact, tuple())]

    def duplicate_hash(self) -> Hashable:
        return self.str_type_id

    @staticmethod
    def _collect_components(props) -> Iterator:
        for p in props:
            if isinstance(p, ComponentFact):
                yield p.str_type_id, p.target

    def choose_one(
            self, other, related: Callable[[str], Iterable[str]], restore_element: Callable[[str], EnsureIdentifiable]
    ) -> Optional[Tuple[str, str]]:
        if not isinstance(other, CompositeValueFact) or self.str_type_id != other.str_type_id:
            return None
        if self.status is FactStatus.APPROVED and other.status is FactStatus.APPROVED:
            return None
        props = set(self._collect_components(map(restore_element, related(self.id))))
        other_props = set(self._collect_components(map(restore_element, related(other.id))))
        if props <= other_props or other_props <= props:
            return (self.id, other.id) if self.status < other.status else (other.id, self.id)
        return None


ValueFact = Union[AtomValueFact, CompositeValueFact]


@generate_model(label='component')
@dataclass(frozen=True, eq=False)
class ComponentFact(Identifiable, AbstractLinkFact[CompositeValueFact, ValueFact, ComponentValueType]):
    """
    Represents a composite value component fact.
    It links composite value fact with another value fact (its component).
    """
    pass



from abc import ABCMeta
from dataclasses import dataclass, replace
from typing import Callable, Generic, Hashable, Iterable, Optional, Set, Tuple, Type, TypeVar, Union

from typing_extensions import Self

from tdm.abstract.datamodel.domain import AbstractDomain, AbstractDomainType, AbstractLinkDomainType
from tdm.abstract.datamodel.duplicatable import Duplicatable
from tdm.abstract.datamodel.identifiable import EnsureIdentifiable
from tdm.helper import generics_mapping, unfold_union
from ._fact import AbstractFact

_ST = TypeVar('_ST', bound=AbstractFact)
_TT = TypeVar('_TT', bound=AbstractFact)
_LDT = TypeVar('_LDT', bound=AbstractLinkDomainType)


@dataclass(frozen=True)
class AbstractLinkFact(AbstractFact, Duplicatable, Generic[_ST, _TT, _LDT], metaclass=ABCMeta):
    """
    Base class for link facts that connect between other facts.

    Attributes
    ----------
    type_id:
        The fact type identifier or instance of the link domain type.
    source:
        The source linked fact.
    target:
        The target linked fact.
    value:
        KB link identifier for approved link fact.

    """
    type_id: Union[str, _LDT]
    source: _ST
    target: _TT
    value: Optional[str] = None

    @property
    def str_type_id(self) -> str:
        return self.type_id if isinstance(self.type_id, str) else self.type_id.id

    def replace_with_domain(self, domain: AbstractDomain) -> Self:
        """
        Create a new link fact with the domain type instead of its identifier.
        This method also validates domain type restrictions.

        :param domain: The domain to be used for fact validation.
        :return: A new link fact object with the domain type instead of its identifier.
        """
        domain_type: Type[_LDT] = generics_mapping(type(self)).get(_LDT)
        type_id = domain.get_type(self.type_id) if isinstance(self.type_id, str) else self.type_id
        if not isinstance(type_id, domain_type):
            raise ValueError

        source = self.source.replace_with_domain(domain)
        target = self.target.replace_with_domain(domain)

        if self.type_id is type_id and self.source is source and self.target is target:
            return self
        return replace(self, type_id=type_id, source=source, target=target)

    def _as_tuple(self) -> tuple:
        return self.id, self.str_type_id, self.source, self.target, self.value

    def duplicate_hash(self) -> Hashable:
        return self.str_type_id, self.source.duplicate_hash(), self.target.duplicate_hash()

    def choose_one(
            self, other, related: Callable[[str], Iterable[str]], restore_element: Callable[[str], EnsureIdentifiable]
    ) -> Optional[Tuple[str, str]]:
        if not isinstance(other, type(self)) or self.str_type_id != other.str_type_id:
            return None
        from tdm.abstract.datamodel import FactStatus
        if self.status is FactStatus.APPROVED and other.status is FactStatus.APPROVED:
            return None
        if self.source != other.source or self.target != other.target:
            return None
        return (self.id, other.id) if self.status < other.status else (other.id, self.id)

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return NotImplemented
        return self._as_tuple() == other._as_tuple()

    def __hash__(self):
        return hash(self._as_tuple())

    def __post_init__(self):
        mapping = generics_mapping(type(self))
        if not isinstance(self.type_id, str):
            domain_type = mapping.get(_LDT)
            if not issubclass(domain_type, AbstractLinkDomainType):
                raise ValueError
            if not isinstance(self.type_id, domain_type):
                raise ValueError(f"illegal type id for fact {self}. {domain_type} is expected")
            self._check_argument(self.type_id.source, self.source.type_id)
            self._check_argument(self.type_id.target, self.target.type_id)
        if not isinstance(self.source, unfold_union(mapping.get(_ST))):
            raise ValueError(f"illegal source fact: {self.source}, {mapping.get(_ST)} is expected")
        if not isinstance(self.target, unfold_union(mapping.get(_TT))):
            raise ValueError(f"illegal target fact: {self.target}, {mapping.get(_TT)} is expected")

    @staticmethod
    def _check_argument(expected_type: AbstractDomainType, actual_type: Union[str, AbstractDomainType]) -> None:
        if isinstance(actual_type, str) and expected_type.id != actual_type:
            raise ValueError
        elif isinstance(actual_type, AbstractDomainType) and expected_type != actual_type:
            raise ValueError

    @classmethod
    def constant_fields(cls) -> Set[str]:
        return {'source', 'target', 'type_id'}

    @staticmethod
    def type_id_filter(type_id: Union[str, Iterable[str]]) -> Callable[['AbstractLinkFact'], bool]:
        """
        Generate a filter function to filter link facts based on their type identifier or a collection of type identifiers.

        :param type_id: The type identifier or collection of type identifiers to filter.
        :return: The filter function that returns True for link facts with a matching type identifier.
        """
        if isinstance(type_id, str):
            def _filter(fact: AbstractLinkFact) -> bool:
                return fact.str_type_id == type_id
        else:
            type_ids = frozenset(type_id)

            def _filter(fact: AbstractLinkFact) -> bool:
                return fact.type_id in type_ids

        return _filter

    @staticmethod
    def source_filter(filter_: Callable[[_ST], bool]) -> Callable[['AbstractLinkFact'], bool]:
        """
        Generate a filter function to filter link facts based on the source fact filter.

        :param filter_: The filter function to apply on the source fact.
        :return: The filter function that returns True for link facts with a matching source fact.
        """

        def _filter(fact: AbstractLinkFact) -> bool:
            return filter_(fact.source)

        return _filter

    @staticmethod
    def target_filter(filter_: Callable[[_TT], bool]) -> Callable[['AbstractLinkFact'], bool]:
        """
        Generate a filter function to filter link facts based on the target fact filter.

        :param filter_: The filter function to apply on the target fact.
        :return: The filter function that returns True for link facts with a matching target fact.
        """

        def _filter(fact: AbstractLinkFact) -> bool:
            return filter_(fact.target)

        return _filter



from abc import ABCMeta
from dataclasses import dataclass, replace
from typing import Callable, Hashable, Iterable, Optional, Set, Tuple, Union

from tdm.abstract.datamodel import AbstractDomain, AbstractFact, AbstractNode, AbstractNodeMention, EnsureIdentifiable, Identifiable
from tdm.abstract.datamodel.duplicatable import Duplicatable
from tdm.abstract.json_schema import generate_model
from .value import AtomValueFact


@dataclass(frozen=True)
class _MentionFact(AbstractFact, metaclass=ABCMeta):
    """
    Auxiliary class for `MentionFact` to fix dataclass fields order.

    Attributes
    ----------
    mention:
        The part of the node that contains value fact mention.
    value:
        The value fact mentioned in document node.
    """
    mention: AbstractNodeMention
    value: AtomValueFact

    @classmethod
    def constant_fields(cls) -> Set[str]:
        return {'mention', 'value'}


@generate_model(label='mention')
@dataclass(frozen=True)
class MentionFact(Identifiable, _MentionFact, Duplicatable):
    """
    Represents fact that some `AtomValueFact` is mentioned in the document node.

    """

    def replace_with_domain(self, domain: AbstractDomain) -> 'MentionFact':
        value = self.value.replace_with_domain(domain)
        if value is self.value:
            return self
        return replace(self, value=value)

    @staticmethod
    def node_filter(node: Union[AbstractNode, str]) -> Callable[['MentionFact'], bool]:
        """
        Create a filter function to filter `MentionFact` instances based on the document node.

        :param node: The document node or its identifier to filter by.
        :return: A filter function for `MentionFact` instances.
        """
        node_id = node.id if isinstance(node, AbstractNode) else node

        def _filter(fact: MentionFact) -> bool:
            return fact.mention.node_id == node_id

        return _filter

    @staticmethod
    def value_filter(filter_: Callable[[AtomValueFact], bool]) -> Callable[['MentionFact'], bool]:
        """
        Create a filter function to filter `MentionFact` instances based on the value.

        :param filter_: A filter function for mentioned `AtomValueFact`.
        :return: A filter function for `MentionFact` instances.
        """

        def _filter(fact: MentionFact) -> bool:
            return filter_(fact.value)

        return _filter

    def duplicate_hash(self) -> Hashable:
        return (
            self.value.duplicate_hash() if isinstance(self.value, Duplicatable) else self.value.id,
            self.mention.duplicate_hash() if isinstance(self.mention, Duplicatable) else self.mention.node_id
        )

    def choose_one(
            self, other, related: Callable[[str], Iterable[str]], restore_element: Callable[[str], EnsureIdentifiable]
    ) -> Optional[Tuple[str, str]]:
        if isinstance(other, MentionFact) and self.mention == other.mention and self.value.id == other.value.id:
            return (self.id, other.id) if self.status < other.status else (other.id, self.id)
        return None



from abc import ABCMeta, abstractmethod
from dataclasses import dataclass
from typing import Callable, Generic, Hashable, Iterable, Iterator, Optional, Sequence, Set, Tuple, TypeVar, Union

from tdm.abstract.datamodel.domain import AbstractValueDomainType
from tdm.abstract.datamodel.duplicatable import Duplicatable
from tdm.abstract.datamodel.identifiable import EnsureIdentifiable
from tdm.abstract.datamodel.value import EnsureConfidenced
from tdm.helper import generics_mapping, is_subclass, unfold_union
from ._fact import AbstractFact, FactStatus

_V = TypeVar('_V', bound=EnsureConfidenced)
_DT = TypeVar('_DT', bound=AbstractValueDomainType)


@dataclass(frozen=True)
class AbstractValueFact(AbstractFact, Duplicatable, Generic[_DT, _V], metaclass=ABCMeta):
    """
    Base class for value facts that represent some KB values.

    If the fact status is FactStatus.NEW `value` should be a tuple.
    If the fact status is FactStatus.APPROVED `value` should be single-valued.

    Attributes
    ----------
    type_id:
        The fact type identifier or instance of the value domain type.
    value:
        The normalized value or tuple of values (sorted by confidence) associated with the value fact.
    """
    type_id: Union[str, _DT]
    value: Union[Tuple[_V, ...], _V] = tuple()

    @property
    def str_type_id(self) -> str:
        return self.type_id if isinstance(self.type_id, str) else self.type_id.id

    @property
    def most_confident_value(self) -> Optional[_V]:
        if not self.value:
            return None
        if isinstance(self.value, Tuple):
            return self.value[0]
        return self.value

    def _tuple_value(self) -> tuple:
        return self.value if isinstance(self.value, tuple) else (self.value,)

    def duplicate_hash(self) -> Hashable:
        return self.str_type_id, self._tuple_value()

    @abstractmethod
    def _collect_links(self, links: Iterable) -> Iterator[Hashable]:
        pass

    def choose_one(
            self, other, related: Callable[[str], Iterable[str]], restore_element: Callable[[str], EnsureIdentifiable]
    ) -> Optional[Tuple[str, str]]:
        if not isinstance(other, type(self)) or self.str_type_id != other.str_type_id:
            return None
        if self.status is FactStatus.APPROVED and other.status is FactStatus.APPROVED:
            return None
        if self._tuple_value() != other._tuple_value():
            return None
        if not self.value:  # equal empty values => check links
            links = set(self._collect_links(map(restore_element, related(self.id))))
            other_links = set(self._collect_links(map(restore_element, related(other.id))))
            if not (links <= other_links or other_links <= links):
                return None
        return (self.id, other.id) if self.status < other.status else (other.id, self.id)

    def __post_init__(self):
        mapping = generics_mapping(type(self))
        domain_type = mapping.get(_DT)
        value_type = mapping.get(_V)
        if not issubclass(domain_type, AbstractValueDomainType):
            raise ValueError
        if not is_subclass(value_type, domain_type.get_value_type()):
            raise ValueError
        if not isinstance(self.type_id, str):
            if not isinstance(self.type_id, domain_type):
                raise ValueError(f"Illegal type id for fact {self}. {domain_type} is expected")
            value_type = self.type_id.value_type
        value_type = unfold_union(value_type)

        if isinstance(self.value, Sequence) and not isinstance(self.value, value_type):  # value could implement Sequence?
            if any(not isinstance(v, value_type) for v in self.value):
                raise ValueError(f"Fact {self} value should be {value_type} or tuple of {value_type}")
            object.__setattr__(self, 'value', tuple(self.value))
        elif not isinstance(self.value, value_type):
            raise ValueError(f"Fact {self} value should be {value_type} or tuple of {value_type}")

        if self.status is FactStatus.NEW and isinstance(self.value, value_type):
            object.__setattr__(self, 'value', (self.value,))
        elif self.status is FactStatus.APPROVED and isinstance(self.value, tuple) and not isinstance(self.value, value_type):
            if len(self.value) != 1 or not isinstance(self.value[0], value_type):
                raise ValueError(f"Approved fact {self} should have single value")
            object.__setattr__(self, 'value', self.value[0])

        if isinstance(self.value, Tuple):
            if all(v.confidence is not None for v in self.value):
                object.__setattr__(self, 'value', tuple(sorted(self.value, key=lambda v: v.confidence, reverse=True)))
            elif not all(v.confidence is None for v in self.value):
                raise ValueError(f"Fact {self} couldn't contain values with confidence and values without confidence at same time")

    @classmethod
    def constant_fields(cls) -> Set[str]:
        return {'type_id'}



import logging
from dataclasses import dataclass, replace
from typing import Callable, Hashable, Iterable, Iterator, Tuple, TypeVar, Union

from typing_extensions import Self

from tdm.abstract.datamodel import AbstractDomain, AbstractFact, AbstractLinkFact, AbstractValue, FactStatus, \
    Identifiable
from tdm.abstract.datamodel.fact import AbstractValueFact
from tdm.abstract.datamodel.value import AbstractConceptValue, EnsureConfidenced
from tdm.abstract.json_schema import generate_model
from tdm.datamodel.common import PrunableMixin, ViewContainer
from tdm.datamodel.domain import PropertyType
from tdm.datamodel.domain.types import AbstractConceptType
from tdm.datamodel.facts.value import ValueFact

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class _KBConceptValue(EnsureConfidenced):
    """
    Auxiliary class for ``KBConceptValue`` to fix dataclass fields order.

    Attributes
    ----------
    concept:
        The concept identifier associated with the fact.
    """
    concept: str


@dataclass(frozen=True)
class KBConceptValue(AbstractConceptValue, _KBConceptValue):
    """
    Represents a concept fact value – concept identifier in the knowledge base.
    """

    @classmethod
    def build(cls, value: Union[str, 'KBConceptValue']) -> Self:
        """
        Build a ConceptValue object from a value.

        :param value: The value to build the KBConceptValue object from.
        :return: The KBConceptValue object.
        """
        if isinstance(value, KBConceptValue):
            return value
        if isinstance(value, str):
            return cls(concept=value)
        raise ValueError


_PT = TypeVar('_PT', bound=PropertyType)
_VT = TypeVar('_VT', bound=AbstractValue)


@dataclass(frozen=True)
class ConceptPropertyValue(object):
    """
    Represents a concept property value.

    Attributes
    ----------
    type_id:
        Either a string or a PropertyType object representing the property type associated with the concept.
    value:
        An AbstractValue object representing the value associated with the concept property.
    """
    type_id: Union[str, PropertyType]
    value: AbstractValue

    @classmethod
    def from_property_fact(cls, fact: 'PropertyFact') -> Self:
        """
        Constructs a ConceptPropertyValue object from a PropertyFact object.

        :param fact: The PropertyFact object to construct the ConceptPropertyValue from.
        :return: The ConceptPropertyValue object
        """
        value = fact.target.value
        if isinstance(value, tuple):
            if len(value) > 1:
                logger.warning(f"fact {fact} contains more than one possible value. Only most confident will be taken into account")
            value = value[0]
        return cls(type_id=fact.type_id, value=value)

    def __eq__(self, other):
        if not isinstance(other, ConceptPropertyValue):
            return NotImplemented
        type_id = self.type_id.id if isinstance(self.type_id, PropertyType) else self.type_id
        other_type_id = other.type_id.id if isinstance(other.type_id, PropertyType) else other.type_id
        return type_id == other_type_id and self.value == other.value

    def __hash__(self):
        return hash((self.type_id, self.value))


@dataclass(frozen=True)
class _MissedConceptValue(EnsureConfidenced):
    """
    Auxiliary class for ``MissedConceptValue`` to fix dataclass fields order.

    Attributes
    ----------
    filters:
        A non-empty tuple of ConceptPropertyValue objects representing the filters associated with the missed value.
    """
    filters: Tuple[ConceptPropertyValue, ...]

    def __post_init__(self):
        if not self.filters:
            raise ValueError


@dataclass(frozen=True)
class MissedConceptValue(AbstractConceptValue, _MissedConceptValue):
    """
    Represents a not-in-kb concept value.
    """
    pass


ConceptValue = Union[KBConceptValue, MissedConceptValue]


@generate_model(label='concept')
@dataclass(frozen=True)
class ConceptFact(Identifiable, AbstractValueFact[AbstractConceptType, ConceptValue], PrunableMixin):
    """
    Represents a fact about some KB concept.

    If the fact status is FactStatus.NEW `value` should be a tuple.
    If the fact status is FactStatus.APPROVED `value` should be single-valued.
    """

    def is_hanging(self, doc: ViewContainer) -> bool:
        return self.status is not FactStatus.APPROVED and \
            not tuple(doc.related_elements(AbstractFact, self, PropertyFact, tuple())) and \
            not self.value

    def __post_init__(self):
        Identifiable.__post_init__(self)
        if isinstance(self.value, tuple):
            if len([v for v in self.value if isinstance(v, MissedConceptValue)]) > 1:
                raise ValueError(f"Concept fact {self} can't contain more than 1 missed value")
        if isinstance(self.value, MissedConceptValue):
            if self.status is FactStatus.APPROVED:
                raise ValueError(f"Approved fact {self} can't be with missed value.")

    def replace_with_domain(self, domain: AbstractDomain) -> 'ConceptFact':
        if isinstance(self.type_id, str):
            domain_type = domain.get_type(self.type_id)
            if not isinstance(domain_type, AbstractConceptType):
                raise ValueError
            return replace(self, type_id=domain_type)
        return self

    def _as_tuple(self) -> tuple:
        return self.id, self.str_type_id, self.value

    def __eq__(self, other):
        if not isinstance(other, ConceptFact):
            return NotImplemented
        return self._as_tuple() == other._as_tuple()

    def __hash__(self):
        return hash(self._as_tuple())

    def _collect_links(self, links: Iterable) -> Iterator[Hashable]:
        for p in links:
            if isinstance(p, PropertyFact):
                yield p.str_type_id, p.target

    @staticmethod
    def empty_value_filter() -> Callable[['ConceptFact'], bool]:
        """
        Create a filter to check if a concept fact has no value.

        :return: The filter function.
        """
        return lambda f: isinstance(f.value, tuple) and not f.value

    @staticmethod
    def tuple_value_filter() -> Callable[['ConceptFact'], bool]:
        """
        Create a filter to check if a concept fact has a tuple value.

        :return: The filter function.
        """
        return lambda f: isinstance(f.value, tuple)

    @staticmethod
    def single_value_filter() -> Callable[['ConceptFact'], bool]:
        """
        Create a filter to check if a concept fact has a single value.

        :return: The filter function.
        """
        return lambda f: isinstance(f.value, ConceptValue)


@generate_model(label='property')
@dataclass(frozen=True, eq=False)
class PropertyFact(Identifiable, AbstractLinkFact[ConceptFact, ValueFact, PropertyType]):
    """
    Represents a concept property fact.
    It links concept fact with some value fact.
    """
    pass



__all__ = [
    'DateTimeValue', 'Date', 'Time',
    'GeoPointValue', 'Coordinates',
    'LinkValue',
    'DoubleValue', 'IntValue', 'StringValue', 'TimestampValue'
]

from .date import Date, DateTimeValue, Time
from .geo import Coordinates, GeoPointValue
from .link import LinkValue
from .scalar import DoubleValue, IntValue, StringValue, TimestampValue



from typing import Dict, Generic, Hashable, Iterator, Set, TypeVar

_K = TypeVar('_K', bound=Hashable)
_V = TypeVar('_V', bound=Hashable)


class LazyUpdatingDict(Generic[_K, _V]):
    def __init__(self, data: Dict[_K, Set[_V]]):
        self._data: Dict[_K, Set[_V]] = dict(data)
        self._updated: Set[_K] = set()

    def extract_data(self) -> Dict[_K, Set[_V]]:
        try:
            return self._data
        finally:
            self._data = None
            self._updated = None

    def _copy_if_not_updated(self, key: _K) -> Set[_V]:
        if key not in self._updated:
            self._data[key] = set(self._data.get(key, set()))
            self._updated.add(key)
        return self._data[key]

    def add(self, key: _K, value: _V) -> None:
        self._copy_if_not_updated(key).add(value)

    def remove(self, key: _K, value: _V) -> None:
        values = self._copy_if_not_updated(key)
        values.discard(value)
        if not values:
            self.pop(key)

    def pop(self, key: _K) -> None:
        self._data.pop(key, None)
        self._updated.discard(key)

    def iterate(self, key: _K) -> Iterator[_V]:
        yield from self._data.get(key, set())

    def len(self, key: _K) -> int:
        return len(self._data.get(key, ()))

    def __contains__(self, key: _K) -> bool:
        return key in self._data



from typing import Dict

from ._view import AbstractView, generate_dataclass_view


def pack_view_state(id2view: Dict[str, AbstractView]) -> dict:
    result = {
        id_: (_orig_type(view), view.__dict__)
        for id_, view in id2view.items()
    }
    return result


def _orig_type(view: AbstractView) -> type:
    if isinstance(view, AbstractView):
        return view.orig_type()
    return type(view)


def unpack_view_state(state: dict) -> Dict[str, AbstractView]:
    result = {
        id_: generate_dataclass_view(cls)[0](**data)
        for id_, (cls, data) in state.items()
    }
    return result



__all__ = [
    'FactStatus',
    'ConceptFact', 'ConceptValue', 'KBConceptValue', 'MissedConceptValue', 'PropertyFact',
    'PropertyFact', 'RelationFact', 'RelationPropertyFact', 'ComponentFact',
    'MentionFact',
    'AtomValueFact', 'CompositeValueFact', 'ValueFact'
]

from tdm.abstract.datamodel import FactStatus
from .concept import ConceptFact, ConceptValue, KBConceptValue, MissedConceptValue, PropertyFact
from .links import RelationFact, RelationPropertyFact
from .mention import MentionFact
from .value import AtomValueFact, ComponentFact, CompositeValueFact, ValueFact



from abc import ABCMeta
from dataclasses import dataclass
from typing import Generic, Set, TypeVar

from tdm.helper import generics_mapping, unfold_union
from ._type import AbstractDomainType

_ST = TypeVar('_ST', bound=AbstractDomainType)
_TT = TypeVar('_TT', bound=AbstractDomainType)


@dataclass(frozen=True)
class AbstractLinkDomainType(AbstractDomainType, Generic[_ST, _TT], metaclass=ABCMeta):
    """
    Base class for knowledge base link types (connecting between other domain types).

    Attributes
    --------
    source:
        The source domain type.
    target:
        The target domain type.
    """
    source: _ST
    target: _TT

    def __post_init__(self):
        types_mapping = generics_mapping(type(self))
        if not isinstance(self.source, unfold_union(types_mapping.get(_ST))):
            raise ValueError(f"Illegal source for {type(self)}. Expected: {types_mapping.get(_ST)}, actual: {type(self.source)}")
        if not isinstance(self.target, unfold_union(types_mapping.get(_TT))):
            raise ValueError(f"Illegal target for {type(self)}. Expected: {types_mapping.get(_TT)}, actual: {type(self.target)}")

    def pretty(self, verbose: bool = False) -> str:
        return f"{type(self).__name__}[{self.id}]({self.name}){{{self.source.pretty(verbose)} --> {self.target.pretty(verbose)}}}"

    @classmethod
    def constant_fields(cls) -> Set[str]:
        return {'source', 'target'}



__all__ = [
    'Domain', 'DomainManager',
    'set_default_domain', 'get_default_domain',
    'AccountType', 'AtomValueType', 'CompositeValueType', 'ConceptType', 'DocumentType', 'PlatformType', 'PropertyType',
    'RelationPropertyType', 'RelationType', 'ComponentValueType'
]

from typing import Callable, Optional

from ._impl import Domain
from ._manager import DomainManager
from .types import AccountType, AtomValueType, ComponentValueType, CompositeValueType, ConceptType, DocumentType, PlatformType, \
    PropertyType, RelationPropertyType, RelationType

DEFAULT_DOMAIN_MANAGER = DomainManager()

set_default_domain: Callable[[Optional[Domain]], None] = DEFAULT_DOMAIN_MANAGER.set
get_default_domain: Callable[[], Domain] = DEFAULT_DOMAIN_MANAGER.get



from dataclasses import dataclass
from typing import Optional, Union

from typing_extensions import Self


@dataclass(frozen=True)
class EnsureConfidenced(object):
    """
    Base interface for objects with confidence that contains non-default fields.
    Python dataclasses before 3.10 couldn't define kwargs-only fields and ``Confidenced`` class contains one (confidence).

    In cases AbstractValue object contains non-default fields one should create intermediate ``class _MyValue(EnsureConfidenced)`` where
    fields are defined.
    Final class will inherit both classes ``class MyValue(Confidenced, _MyValue)``
    """

    def __post_init__(self):
        if not isinstance(self, Confidenced):
            raise TypeError(f"{type(self)} should inherit {AbstractValue}. Actual mro is {type(self).mro()}")


@dataclass(frozen=True)
class Confidenced(object):
    """
    Base class for values with confidence.

    Attributes
    ----------
    confidence:
        Confidence value associated with the value.
    """
    confidence: Optional[Union[float, int]] = None

    def __post_init__(self):
        if self.confidence is not None and not 0 < self.confidence <= 1:
            raise ValueError(f"value confidence should be in interval (0; 1], {self.confidence} is given")
        for type_ in type(self).mro():
            if issubclass(type_, Confidenced):
                continue
            if hasattr(type_, '__post_init__'):
                type_.__post_init__(self)


@dataclass(frozen=True)
class AbstractConceptValue(Confidenced):
    """
    Base class for values containing KB concept identifiers
    """
    pass


@dataclass(frozen=True)
class AbstractValue(Confidenced):
    """
    Base class for normalized values in the knowledge base.
    """

    @classmethod
    def from_dict(cls, value: dict) -> Self:
        """
        Create an instance of value from a dictionary.

        :param value: Dictionary representing the value.
        :return: An instance of the value class created from the dictionary.
        """
        return cls(**value)



__all__ = [
    'ViewContainer',
    'TypedIdsContainer',
    'AbstractView', 'restore_object', 'PrunableMixin'
]

from ._container import TypedIdsContainer
from ._impl import PrunableMixin, ViewContainer
from ._view import AbstractView, restore_object



from abc import abstractmethod
from typing import Generic, TypeVar

from typing_extensions import Self

from tdm.abstract.datamodel import AbstractNode

_Node = TypeVar('_Node', bound=AbstractNode)


class AbstractNodeWrapper(AbstractNode, Generic[_Node]):
    """
    Abstract class for node wrappers inheriting from AbstractNode.

    Use `tdm.wrapper.node.generate_wrapper` decorator to generate implementation

    """

    @classmethod
    @abstractmethod
    def wrap(cls, node: _Node) -> Self:
        """
        Wrap the node.

        :param node: The node to be wrapped.
        :return: A wrapped node.
        """
        pass



from typing import Any, Dict, Generic, List, Mapping, Set, Type, TypeVar

from .abstract import AbstractElementSerializer

_SEQUENCES: Mapping[type, type] = {
    tuple: List,  # in other case Tuple[A, ...] is converted to Tuple[A]. Deserializes to tuple
    set: Set,
    list: List
}


class SequenceElementSerializer(AbstractElementSerializer):
    def __init__(self, real_type: type, serializer: AbstractElementSerializer):
        self._real_type = real_type
        self._typing_type = _SEQUENCES[real_type]
        self._serializer = serializer

    def serialize(self, element):
        return self._real_type(self._serializer.serialize(e) for e in element)

    def deserialize(self, serialized, typed_id2element: Dict[type, Dict[str, Any]]):
        return self._real_type(self._serializer.deserialize(s, typed_id2element) for s in serialized)

    def field_type(self, element_type):
        return self._typing_type[element_type]


_E = TypeVar('_E')
_S = TypeVar('_S')


class MappingElementSerializer(AbstractElementSerializer[Mapping[str, _E], Dict[str, _S]], Generic[_E, _S]):
    def __init__(self, real_type: type, serializer: AbstractElementSerializer[_E, _S]):
        self._real_type = real_type
        self._serializer = serializer

    def serialize(self, element: Mapping[str, _E]) -> Dict[str, _S]:
        return {key: self._serializer.serialize(value) for key, value in element.items()}

    def deserialize(self, serialized: Dict[str, _S], typed_id2element: Dict[type, Dict[str, Any]]) -> Mapping[str, _E]:
        return {key: self._serializer.deserialize(value, typed_id2element) for key, value in serialized.items()}

    def field_type(self, element_type) -> Type[Dict[str, _S]]:
        return Dict[str, element_type]



from typing import Any, Dict, Type

from tdm.abstract.datamodel import AbstractMarkup, FrozenMarkup
from tdm.helper import unfreeze
from .abstract import AbstractElementSerializer


class MarkupSerializer(AbstractElementSerializer[AbstractMarkup, dict]):
    def serialize(self, element: AbstractMarkup) -> dict:
        return unfreeze(element.markup)

    def deserialize(self, serialized: dict, typed_id2element: Dict[type, Dict[str, Any]]) -> FrozenMarkup:
        return FrozenMarkup.freeze(serialized)

    def field_type(self, element_type: Type[AbstractMarkup]) -> Type[dict]:
        return dict



from dataclasses import replace
from typing import Callable, Iterable


def getter_delegate(name: str, post_processors: Iterable[Callable] = tuple(), *, base: str = 'markup'):
    def f(self, *args, **kwargs):
        result = getattr(getattr(self, base), name)(*args, **kwargs)
        for post in post_processors:
            result = post(self, result)
        return result

    return f


def property_delegate(name: str, post_processors: Iterable[Callable] = tuple(), *, base: str = 'markup'):
    def f(self):
        result = getattr(getattr(self, base), name)
        for post in post_processors:
            result = post(self, result)
        return result

    return property(f)


def modifier_delegate(name: str, validators: Iterable[Callable] = tuple(), *, base: str = 'markup'):
    def f(self, *args, **kwargs):
        for validator in validators:
            ret = validator(self, *args, **kwargs)
            if ret is not None:
                args = ()
                kwargs = ret
        return replace(self, **{base: getattr(getattr(self, base), name)(*args, **kwargs)})

    return f



from dataclasses import dataclass
from typing import Union

from tdm.abstract.datamodel import AbstractLinkDomainType, Identifiable
from ._composite import CompositeValueType
from ._concept import AbstractConceptType
from ._relation import RelationType
from ._value import AtomValueType

ValueType = Union[CompositeValueType, AtomValueType]


@dataclass(frozen=True)
class PropertyType(Identifiable, AbstractLinkDomainType[AbstractConceptType, ValueType]):
    """
    Domain type for property facts
    """

    def __hash__(self):
        return hash(self.id)


@dataclass(frozen=True)
class RelationPropertyType(Identifiable, AbstractLinkDomainType[RelationType, ValueType]):
    """
    Domain type for relation property facts
    """
    pass



from abc import abstractmethod
from collections import defaultdict
from itertools import chain
from operator import itemgetter
from typing import Callable, Dict, Hashable, Iterable, Iterator, List, Optional, Set, Tuple, Type, TypeVar, Union

from typing_extensions import Self

from tdm.abstract.datamodel import AbstractFact, EnsureIdentifiable, Identifiable, and_filter
from tdm.abstract.datamodel.duplicatable import Duplicatable
from tdm.helper import unfold_union
from ._container import TypedIdsContainer
from ._state import pack_view_state, unpack_view_state
from ._types import get_base_type
from ._util import LazyUpdatingDict
from ._view import AbstractView, object_view, restore_object

_I = TypeVar('_I', bound=Identifiable)
_Base = TypeVar('_Base', bound=EnsureIdentifiable)


class ViewContainer(object):
    __slots__ = (
        '_id2view', '_dependencies', '_containers', '_hash2ids', '_id2hash', '_duplicate2id', '_id2duplicates'
    )

    def __setstate__(self, state):
        d, s = state
        if d is not None:
            self.__dict__.update(d)
        s['_id2view'] = unpack_view_state(s['_id2view'])
        for component in chain.from_iterable(getattr(cls, '__slots__', ()) for cls in type(self).mro()):
            setattr(self, component, s.get(component, None))

    def __getstate__(self):
        result = {component: getattr(self, component, None) for component in
                  chain.from_iterable(getattr(cls, '__slots__', ()) for cls in type(self).mro())}
        result['_id2view'] = pack_view_state(result.get('_id2view', {}))
        d = self.__dict__ if hasattr(self, '__dict__') else None
        return d, result

    def __init__(
            self,
            id2view: Dict[str, Union[EnsureIdentifiable, AbstractView]],
            dependencies: Dict[str, Set[Tuple[str, Type[EnsureIdentifiable]]]],
            containers: Dict[Type[EnsureIdentifiable], TypedIdsContainer],
            hash2ids: Dict[Hashable, Set[str]],
            id2hash: Dict[str, Hashable],
            duplicate2id: Dict[str, str],
            id2duplicates: Dict[str, Set[str]]
    ):
        self._id2view = id2view
        self._dependencies = dependencies
        self._containers = containers
        self._hash2ids = hash2ids
        self._id2hash = id2hash
        self._duplicate2id = duplicate2id
        self._id2duplicates = id2duplicates

    def _replace(self, **kwargs) -> Self:
        return type(self)(
            **{
                'id2view': self._id2view,
                'dependencies': self._dependencies,
                'containers': self._containers,
                'hash2ids': self._hash2ids,
                'id2hash': self._id2hash,
                'duplicate2id': self._duplicate2id,
                'id2duplicates': self._id2duplicates,
                **kwargs
            }
        )

    def id2element(self, base_type: Type[_Base]) -> Dict[str, _Base]:
        return {i: restore_object(self._id2view[self._duplicate2id.get(i, i)], self._id2view) for i, _ in self._containers[base_type]}

    def elements(self, base_type: Type[_Base]) -> Dict[Type[_Base], Iterable[_Base]]:
        return {
            t: {restore_object(self._id2view[i], self._id2view) for i in ids}
            for t, ids in self._containers[base_type].type2ids.items()
        }

    def get_element(self, base_type: Type[EnsureIdentifiable], id_: str) -> _I:
        id_ = self._duplicate2id.get(id_, id_)
        if id_ not in self._containers[base_type]:
            raise KeyError
        return restore_object(self._id2view[id_], self._id2view)

    def get_elements(
            self, base_type: Type[EnsureIdentifiable], type_: Type[_I],
            filter_: Union[Callable[[_I], bool], Iterable[Callable[[_I], bool]]]
    ) -> Iterator[_I]:
        if isinstance(filter_, Iterable):
            filter_ = and_filter(*filter_)

        type_ = unfold_union(type_)
        if not all(issubclass(t, base_type) for t in type_):
            raise ValueError

        for t, ids in self._containers[base_type].type2ids.items():
            if not issubclass(t, type_):
                continue
            yield from filter(filter_, (restore_object(self._id2view[id_], self._id2view) for id_ in ids))

    def related_elements(
            self, base_type: Type[EnsureIdentifiable], obj: Union[Identifiable, str], type_: Type[_I],
            filter_: Union[Callable[[_I], bool], Iterable[Callable[[_I], bool]]]
    ) -> Iterator[_I]:
        if isinstance(obj, Identifiable):
            obj = obj.id
        if not isinstance(obj, str):
            raise ValueError

        if isinstance(filter_, Iterable):
            filter_ = and_filter(*filter_)

        type_ = unfold_union(type_)
        if not all(issubclass(t, base_type) for t in type_):
            raise ValueError

        for element_id, element_type in self._dependencies.get(obj, tuple()):
            if not issubclass(element_type, base_type):
                continue
            view = self._id2view[element_id]
            if not issubclass(view.orig_type(), type_):
                continue
            fact = restore_object(view, self._id2view)
            if filter_(fact):
                yield fact

    @staticmethod
    def _check_duplicate(element, hash2ids, id2hash, id2view, dependencies) -> Optional[Tuple[str, str]]:
        element_hash = id2hash[element.id]
        for candidate_id in hash2ids.iterate(element_hash):
            if candidate_id == element.id:
                continue
            candidate = restore_object(id2view[candidate_id], id2view)
            duplicate = element.choose_one(
                candidate,
                lambda i: map(itemgetter(0), dependencies.iterate(i)),
                lambda i: restore_object(id2view[i], id2view)
            )
            if duplicate is not None:
                return duplicate
        return None

    def with_elements(self, elements: Iterable[EnsureIdentifiable], *, update: bool = False) -> Self:
        containers: Dict[Type[EnsureIdentifiable], TypedIdsContainer] = dict(self._containers)
        id2view = dict(self._id2view)

        id2hash = dict(self._id2hash)
        hash2ids = LazyUpdatingDict(self._hash2ids)

        duplicate2id = dict(self._duplicate2id)
        id2duplicates = LazyUpdatingDict(self._id2duplicates)

        dependencies = LazyUpdatingDict(self._dependencies)

        possible_duplicates: Set[Duplicatable] = set()

        elements = self._order_dependencies(set(elements), update)
        prunable: List[PrunableMixin] = []

        for element_type, element, element_view in elements:
            assert isinstance(element_view, AbstractView)
            if element.id in duplicate2id:  # actually this element should silently overwrite duplicated identifier
                id_ = duplicate2id.pop(element.id)
                id2duplicates.remove(id_, element.id)
            if isinstance(element, PrunableMixin):
                prunable.append(element)
            container = containers[element_type]
            if element.id in container:
                # remove element from duplicate hash store. it should be rehashed as some features could be changed
                if isinstance(element, Duplicatable):
                    h = id2hash.pop(element.id)
                    hash2ids.remove(h, element.id)
                try:
                    id2view[element.id].validate_update(element_view)
                except ValueError as e:
                    raise ValueError(f"Couldn't update element {element}") from e
            elif element.id in id2view:
                raise ValueError(f"Element {element} identifiers collision: document already "
                                 f"contains {id2view[element.id].orig_type()} with same id")

            if isinstance(element, Duplicatable):
                element_hash = element.duplicate_hash()
                if element_hash in hash2ids:  # there are some similar objects
                    possible_duplicates.add(element)
                id2hash[element.id] = element_hash
                hash2ids.add(element_hash, element.id)
                if isinstance(element_view, AbstractView):
                    for dep_id in element_view.__depends_on__:
                        if dep_id not in id2hash:
                            continue
                        dep_hash = id2hash[dep_id]
                        if hash2ids.len(dep_hash) > 1:
                            possible_duplicates.add(restore_object(id2view[dep_id], id2view))

            if isinstance(element_view, AbstractView):
                for dep in element_view.__depends_on__:
                    dependencies.add(dep, (element.id, element_type))
            id2view[element.id] = element_view
            containers[element_type] = containers[element_type].with_ids([(element.id, type(element))])

        # now lets try to resolve duplicates
        while possible_duplicates:
            next_turn = set()
            for element in possible_duplicates:
                if element.id not in id2hash:
                    continue  # already removed
                duplicate = self._check_duplicate(element, hash2ids, id2hash, id2view, dependencies)
                if duplicate is None:
                    continue
                actual_id, obsolete_id = duplicate
                duplicate2id[obsolete_id] = actual_id
                id2duplicates.add(actual_id, obsolete_id)
                # cleanup obsolete element
                # - remove it view
                obsolete_view = id2view.pop(obsolete_id)  # remove view
                base_type = get_base_type(obsolete_view.orig_type())
                # - remove dependencies reverse links
                for dep in obsolete_view.__depends_on__:
                    dependencies.remove(dep, (obsolete_id, base_type))
                for dep_id, bt in dependencies.iterate(obsolete_id):
                    # get elements that depends on duplicate to force update dependencies
                    obj = restore_object(id2view[dep_id].substitute_id(obsolete_id, actual_id), id2view)
                    updated_view = object_view(obj)
                    id2view[dep_id] = updated_view
                    dependencies.add(actual_id, (dep_id, bt))
                    if dep_id in id2hash:  # rehash duplicatable dependencies
                        dep_hash = id2hash[dep_id]
                        hash2ids.remove(dep_hash, dep_id)
                        obj_hash = obj.duplicate_hash()
                        id2hash[dep_id] = obj_hash
                        hash2ids.add(obj_hash, dep_id)
                        next_turn.add(obj)
                        for second_dep in updated_view.__depends_on__:
                            # we should check nodes after links. when values merge, value <- property -> concept should be checked
                            if second_dep != actual_id:  # no need to check same object once again
                                next_turn.add(restore_object(id2view[second_dep], id2view))
                # remove from identifier container
                containers[base_type] = containers[base_type].without_ids([obsolete_id])
                # remove from hash store
                obsolete_hash = id2hash.pop(obsolete_id)
                hash2ids.remove(obsolete_hash, obsolete_id)
            possible_duplicates = next_turn

        res = self._replace(
            id2view=id2view or self._id2view,
            dependencies=dependencies.extract_data(),
            containers=containers or self._containers,
            id2hash=id2hash or self._id2hash,
            hash2ids=hash2ids.extract_data(),
            id2duplicates=id2duplicates.extract_data(),
            duplicate2id=duplicate2id
        )

        pruned = False
        while not pruned:
            res, prunable, pruned = self.prune(res, prunable)

        return res

    @staticmethod
    def prune(res: 'ViewContainer', facts: Iterable['PrunableMixin']) -> Tuple['ViewContainer', Iterable['PrunableMixin'], bool]:
        hanging_ids = set()
        non_pruned_facts = []
        try:
            for element in facts:
                if isinstance(element, PrunableMixin) and (element.is_hanging(res)):
                    hanging_ids.add(element.id)
                else:
                    non_pruned_facts.append(element)
        except KeyError:  # some documents do not have AbstractFact in their .containers at all
            return res, [], True

        if hanging_ids:
            return res.without_elements(hanging_ids, cascade=True), non_pruned_facts, False
        return res, [], True

    def without_elements(self, ids: Iterable[str], *, cascade: bool = False) -> Self:
        ids = set(self._id2view.keys()).intersection(map(_to_id, ids))  # ignore excess ids
        if not cascade:
            # check no hang links
            for id_ in ids:
                for dep, _ in self._dependencies.get(id_, ()):
                    if dep not in ids:
                        raise ValueError(f"Couldn't remove element {id_} as it depends on element {dep}")
        remove_from_containers = defaultdict(set)
        id2view = dict(self._id2view)

        id2hash = dict(self._id2hash)
        hash2ids = LazyUpdatingDict(self._hash2ids)

        duplicate2id = dict(self._duplicate2id)
        id2duplicates = LazyUpdatingDict(self._id2duplicates)

        dependencies = LazyUpdatingDict(self._dependencies)

        remove = defaultdict(set)
        removed_ids = set()

        recheck_ids = set()

        for id_ in ids:
            remove[get_base_type(id2view[id_].orig_type())].add(id_)
            for i in id2duplicates.iterate(id_):
                duplicate2id.pop(i, None)
            id2duplicates.pop(id_)
        while remove:
            to_remove = defaultdict(set)
            for base_type, ids in remove.items():
                for id_ in ids:
                    if id_ in removed_ids:
                        continue
                    removed_ids.add(id_)
                    for dep_id, dep_type in dependencies.iterate(id_):
                        to_remove[dep_type].add(dep_id)
                    view = id2view.pop(id_)
                    if issubclass(view.orig_type(), Duplicatable):
                        # remove element from duplicate hash store
                        duplicate_hash = id2hash.pop(id_)
                        hash2ids.remove(duplicate_hash, id_)
                    dependencies.pop(id_)
                    remove_from_containers[base_type].add(id_)
                    if not isinstance(view, AbstractView):
                        continue
                    for dep in view.__depends_on__:
                        if dep in id2hash:  # it is duplicable
                            recheck_ids.add(dep)
                        if dep in dependencies:
                            dependencies.remove(dep, (id_, base_type))

            remove = to_remove

        containers = dict(self._containers)
        for base_type, ids in remove_from_containers.items():
            containers[base_type] = containers[base_type].without_ids(ids)

        res = self._replace(
            id2view=id2view,
            dependencies=dependencies.extract_data(),
            containers=containers,
            id2hash=id2hash,
            hash2ids=hash2ids.extract_data()
        )

        pruned = False
        facts = list(chain(*(res.elements(AbstractFact).values())))
        while not pruned:
            res, facts, pruned = self.prune(res, facts)

        # FIXME: extremely inefficient solution. We should check duplicates without useless elements updating
        recheck_duplicates = {
            restore_object(res._id2view[id_], res._id2view) for id_ in recheck_ids if id_ in res._id2view
        }
        if recheck_duplicates:
            return res.with_elements(recheck_duplicates)
        return res

    def _order_dependencies(
            self, elements: Set[EnsureIdentifiable], update: bool = False
    ) -> Iterable[Tuple[Type[EnsureIdentifiable], EnsureIdentifiable, AbstractView]]:
        result = []
        visited = set()
        while elements:
            to_process = set()
            for element in map(self._transform_element, elements):
                visited.add(element)
                view = object_view(element)

                result.append((get_base_type(element), element, view))

                if not isinstance(view, AbstractView):
                    continue  # no dependencies

                dependencies = view.get_dependencies(element)
                dependencies.difference_update(elements)  # will be processed next steps
                dependencies.difference_update(visited)
                if update:
                    to_process.update(dependencies)  # add to queue
                else:
                    self._validate_elements(dependencies)
            elements = to_process
        return result[::-1]

    def _transform_element(self, element: _Base) -> _Base:
        return element

    def _validate_elements(self, elements: Set[EnsureIdentifiable]) -> None:
        for element in elements:
            if element.id not in self._id2view:
                raise ValueError(f"document contains no {element}")
            view = self._id2view[element.id]
            element_type = type(element)
            view_type = view.orig_type() if isinstance(view, AbstractView) else type(view)
            if not issubclass(element_type, view_type) or not issubclass(view_type, element_type):
                raise ValueError(f"Type mismatch for {element}. Expected: {view_type}, actual: {element_type}")


def _to_id(obj: Union[str, EnsureIdentifiable]) -> str:
    return obj if isinstance(obj, str) else obj.id


class PrunableMixin:
    @abstractmethod
    def is_hanging(self, doc: ViewContainer) -> bool:
        pass



from typing import Any, Dict, Iterable, Union

from .abstract import AbstractElementSerializer


class CompositeElementSerializer(AbstractElementSerializer):
    def __init__(self, serializers: Iterable[AbstractElementSerializer]):
        self._serializers = tuple(serializers)

    def serialize(self, element):
        for serializer in self._serializers:
            try:
                return serializer.serialize(element)
            except Exception:
                pass
        raise ValueError

    def deserialize(self, serialized, typed_id2element: Dict[type, Dict[str, Any]]):
        for serializer in self._serializers:
            try:
                return serializer.deserialize(serialized, typed_id2element)
            except Exception:
                pass
        raise ValueError

    def field_type(self, element_type):
        types = tuple(serializer.field_type(element_type) for serializer in self._serializers)
        return Union[types]

    @classmethod
    def build(cls, serializers: Iterable[AbstractElementSerializer]) -> AbstractElementSerializer:
        serializers = tuple(serializers)
        if len(serializers) == 1:
            return serializers[0]
        return CompositeElementSerializer(serializers)



from typing import Any, Callable, Optional


def cache_result(key_param: Optional[str] = None, cache: Optional[dict] = None):
    if cache is None:
        cache = {}

    def wrap(f: Callable):
        def wrapper(*args, **kwargs) -> Any:
            if key_param is not None:
                key = kwargs[key_param]
            elif args:
                key = args[0]
            else:
                key = kwargs[next(iter(kwargs))]
            if key in cache:
                return cache[key]
            result = f(*args, **kwargs)
            cache[key] = result
            return result

        return wrapper

    return wrap



from dataclasses import dataclass
from typing import Set

from typing_extensions import Self

from tdm.abstract.datamodel import AbstractLinkDomainType, Identifiable
from ._concept import AbstractConceptType


@dataclass(frozen=True)
class _RelationType(AbstractLinkDomainType[AbstractConceptType, AbstractConceptType]):
    """
    Auxiliary class for ``RelationType`` to fix dataclass fields order.

    Attributes
    ----------
    directed:
        indication of ordered link
    """
    directed: bool = True

    @classmethod
    def constant_fields(cls) -> Set[str]:
        return super().constant_fields().union(('directed',))

    pass


@dataclass(frozen=True)
class RelationType(Identifiable, _RelationType):
    """
    Domain type for relation facts.
    """

    def pretty(self, verbose: bool = False) -> str:
        arrow = '-->' if self.directed else '<->'
        return f"{type(self).__name__}[{self.id}]({self.name}){{{self.source.pretty(verbose)} {arrow} {self.target.pretty(verbose)}}}"

    def inversed(self) -> Self:
        return type(self)(
            **{
                **self.__dict__,
                'source': self.target,
                'target': self.source
            }
        )



from dataclasses import dataclass

from tdm.abstract.datamodel import AbstractDomainType, Identifiable


@dataclass(frozen=True)
class CompositeValueType(Identifiable, AbstractDomainType):
    """
    Domain type for composite value facts.
    """
    pass



from abc import ABCMeta
from dataclasses import dataclass
from typing import Generic, Set, Type, TypeVar

from tdm.abstract.datamodel.value import EnsureConfidenced
from tdm.helper import generics_mapping
from ._type import AbstractDomainType

_VT = TypeVar('_VT', bound=EnsureConfidenced)


@dataclass(frozen=True)
class AbstractValueDomainType(AbstractDomainType, Generic[_VT], metaclass=ABCMeta):
    """
    Base class for knowledge base value-associated types

    Attributes
    --------
    value_type:
        Type of the associated value
    """
    value_type: Type[_VT]

    def pretty(self, verbose: bool = False) -> str:
        if verbose:
            return f"{type(self).__name__}{{{self.value_type.__name__}}}[{self.id}]({self.name})"
        return f"{type(self).__name__}[{self.id}]({self.name})"

    @classmethod
    def constant_fields(cls) -> Set[str]:
        return {'value_type'}

    @classmethod
    def get_value_type(cls) -> type:
        return generics_mapping(cls)[_VT]
